/**** Microsoft C v4.0  msc/Od hash; link/STACK:8192 hash; *****/
/*
*  CLASS     : 245-01 Data Structures
*  INSTRUCTOR: Jim Hein
*  STUDENT   : Michael M. McDaniel
*
*  This Program: HASH.C
*
*  PURPOSE: This program tests various hashing strategies. Input data consists of three letter month
*           abbreviations. Since we want to test various strategies, this program has been implemented
*           as a multi-dimensional array.
*
  This program is set up as a multidimensional array of entries. Each table
  is constructed as needed for its particularly TEST case. Where each key is
  finally located is of course dependant on the hash functions and collision
  reconciliation used for each table. The same initial hash function is used
  for each entry, so the initial hash value is the same in each table.
  However, as can be verified upon final output, the final locations of each
  element is different for the different collision strategies implemented.

  Each page has capabilities of holding  MAX_TABLE (100 in this case) number
  of entries. The user sets the exact number of table entries allowable when
  they answer the beginning prompt for the program.

  Basically, each table can be thought of as just a 'page' to deal with.
  Each 'page' also has its own set of variables to use so there is no
  conflict between pages.

  In order to use the various pages, the following syntax is used: table[0]...   accesses page 0
                                                                   table[1]...   accesses page 1
                                                                   table[2]...   accesses page 2
  If more hashing strategies need to be considered, simply increase NUM_TABLES
  and add the new strategy as a 'case' in the function test()
                                                                   
                    ____________________
                   /                   /|
                  /                   / |
                 /___________________/  |
                /                   /|  |
               /                   / |  |
              /___________________/  |  |
             /                   /|  |  |
            /                   / |  |  |
           /___________________/  |  |  |
           |                   |  |  |  /
           |                   |  |  | / <== page 2  (table for TEST 3)
           |                   |  |  |/
           |                   |  |  /
           |                   |  | / <== page 1  (table for TEST 2)
           |                   |  |/
           |                   |  /
           |                   | / <== page 0  (table for TEST 1 & 4)
           |___________________|/


*  BASIC ALGORITHM:
*  1) set up user variables (size of table and number of keys; size will be modulus operator for functions, also)
*  2) accept input
*     a) convert to upper case
*  3) calculate hash function on input
*  4) place in table; if collision do 5) else goto 2) if input not exhausted
*  5) reconcile collision
*
*  ASSUMPTIONS: Input is correct! Moderate error checking is done, but this is not production quality.
*               Redundant data is accounted for, and all counts are kept correct
*
*  NOTE: This looks much more complicated than it is. It could have been done
*        as three or four separate programs but there's very little real difference between
*        them. Also the information could have been output after each calculation. The decision
*        against this was that the information would have been too hard to analyze as to what
*        was actually happening and which strategies were superior. New hashing and collision
*        strategies can be added to this program relatively easily, and different table sizes
*        or number of keys are strictly a user option at run time.
*
*/

#include <stdio.h>    /* standard i/o compiler package */

   /* NOTE!!! the hash.h file is at the end of this listing in an #ifdef */

#include "hash.h"    /*  structure of information 'nodes' of hash_type */
                    /* includes enough fields so that each test can */
                   /* use the same hash_type structure */

#define newline "\n"
#define nulval  (-1)
#define FALSE   (0)
#define TRUE    (1)

#define mod      %          /* for ye of little 'C' knowledge */
#define MAX_TABLE (100)    /* maximum table size */
#define NUM_TESTS (3)     /* number of different pages of tables  */
                         /* we will construct a multi-dimensional array of NUM_TESTS by table_size */

/***********************  start main program *******************************/
 
main()
{
 int hash();               /* parms(ptr); char *ptr to beginning of 3 character array */
 int lin_probe();         /* parms(place, modulus); int place, modulus */
 void instructions();    /* give user instructions */
 void print_table();    /* parms(page, size, table, keys); prints output table */
                       /* int page size, keys; hash_type table[][] */
 void test();         /* parms(page, place, table, input, size, index); int page, place[], size, index; hash_type table[][], input[] */
 void new_spot();    /* parms(page, place, table, input, size); int page, size, place[]; hash_type table[][], input[] */
 int table_size, num_keys;
 int i, j, k, temp;
 int entries;            /* number of input entries */
 int m, p, r, s;
 int correction[NUM_TESTS];    /* used to correct collision count if redundant data is entered */
 short dont_use[NUM_TESTS];   /* boolean: input exists, so don't use it */
 short check[NUM_TESTS];     /* boolean: check if couldn't put value in table */
 int place[NUM_TESTS];      /* in respective table */
 hash_type input[NUM_TESTS];
 hash_type table[NUM_TESTS][MAX_TABLE];       /* user can play with a table up to this size */

 instructions();      /* give user instructions */

 printf("\nInput will be three letter month abbreviations.\n");
 printf("Maximum table size you can use is %d\n", MAX_TABLE);
 printf("NOTE: that your chosen table size will be your modulus for hash calculations.\n");
 printf("What table size do you want to experiment with? ===> ");
 if(scanf("%d", &table_size) != 1) printf("INPUT ERROR\n");
 if(table_size > MAX_TABLE)
  {
   printf("Sorry, table size choice is over %d; limitting your choice to %d\n", MAX_TABLE, MAX_TABLE);
   table_size = MAX_TABLE;
  }

 printf("How many keys will you have? ===> ");
 if(scanf("%d", &num_keys) != 1) printf("INPUT ERROR\n");
 if(num_keys > MAX_TABLE)
  {
   printf("Sorry, your number of keys choice is over %d; limitting your choice to %d\n", MAX_TABLE, MAX_TABLE);
   num_keys = MAX_TABLE;
  }
 
 printf("\n\n   ===> now initializing table of size %d and type hash_type for %d pages <===\n", table_size, NUM_TESTS);
 printf("                 ===> input will be accepted for %d keys <===\n\n", num_keys);
                  
 for(i=0; i<NUM_TESTS; i++)                   /* initialize tables & variables*/
  {
   input[i].key      = input[i].array;      /* these are */
   sprintf(input[i].key, "%s", "   \0");   /* the variables */
   input[i].link     = nulval;            /* link to next */
   input[i].crash    = 0;                /* number of collisions */
   input[i].hash     = 0;               /* hash value */
   input[i].doubl    = 0;              /* 2nd hash value */
   input[i].lin_prob = 0;             /* linear probe value */
   dont_use[i]       = FALSE;        /* use everything */

  for(j=0; j<MAX_TABLE; j++)        /* and these are the tables */
   {
    table[i][j].key    = table[i][j].array;         /* assign string ptr to array addr */
    sprintf(table[i][j].key, "%s", "   \0");       /* string is empty */
    table[i][j].link     = nulval;                /* all table slots are empty */
    table[i][j].crash    = 0;                    /* initialize */
    table[i][j].hash     = 0;                   /* the rest */
    table[i][j].doubl    = nulval;             /* of this */
    table[i][j].lin_prob = nulval;            /* stuff, too */
   }
  }

 printf("\n\nReady for input; enter your first month now...");
 for(entries=0; entries<num_keys; entries++)
  {
   if(scanf("%s", input[0].key) != 1) printf("INPUT ERROR\n");
    else
     {
      sprintf(input[0].key, "%s", strupr(input[0].key)); /* make upper case */
      place[0]          = hash(input[0].key, table_size);
      input[0].hash     = place[0];

      for(i=1; i<NUM_TESTS; i++)
       {
        sprintf(input[i].key, "%s", input[0].key);    /* all of 'em (for each table) upper case */
        place[i]          = place[0];                /* each TEST starts with the same place value */
        input[i].hash     = place[i];               /* derived from the same first hash value */
        input[i].doubl    = 0;
        input[i].lin_prob = 0;
        correction[i]     = 1;
        check[i]          = FALSE;
       }

   for(i=0; i<NUM_TESTS; i++)
    {
      j = 0;
      while( (check[i] = collision(table[i][place[i]]) == TRUE ) && (++j < table_size) )         /* PERFORM TEST i */
       {
        if( strcmp(table[i][place[i]].key, input[i].key) == 0 )         /* strings match */
         {
          printf("%s already exists in table %d, input disregarded!\n", input[i].key, i);
          dont_use[i] = TRUE;
          if(table[i][place[i]].crash != 0) table[i][place[i]].crash -= correction[i];
          break;              /* break out, don't use this input */
         }
         else
          test(i, place, table, input, table_size, j), correction[i] += 1;                   /* for collision reconciliation */
        }

      if( (j >= table_size) && (check[i] == TRUE) ) printf("\nFATAL ERROR: %s is homeless! in TEST %d (page %d)", input[0].key, (i+1), i);
       else
        if(dont_use[i] == FALSE)  new_spot(i, place, table, input);
         else dont_use[i] = FALSE;
    } /* end for */

     if(entries < (num_keys-1)) printf("\nenter another month now... ");   /* ask for more input */
    } /* end else */
 }   /* end for */

  m = table_size;  /* only for compactness in the print formatting */

  printf(newline);
  for(i=0; i<20; i++) printf("****");
  printf("\n\nTEST 1: Open addressing with linear probing\n");
  printf("hash function: h(xyz) = (ord(x) + ord(z)) mod %d\n", table_size);
  printf("linear probe : (i+1) mod %d, (i+2) mod %d, ... (i+ %d) mod %d\n", m, m, m, m);
  printf("NOTE: that the first hash function value is the first choice of where\n");
  printf("to put the items in the table. For this test, the linear probe value is the\n");
  printf("second choice if there is a collision.\n");
  print_table(0, table_size, table, num_keys);                /* print table 1 */

  printf(newline);
  for(j=0; j<20; j++) printf("****");
  printf("\n\nTEST 2: Open addressing with double hashing\n");
  printf("hash function: h(xyz) = (ord(x) + ord(z)) mod %d\n", table_size);
  printf("2nd hash function: g(xyz) = (ord(x) + ord(y)) mod %d\n", table_size);
  printf("probe function: (i+g(xyz) mod %d, (i+2*g(xyz)) mod %d, ... (i+%d*g(xyz)) mod %d\n", m, m, m, m);
  printf("NOTE: that the first hash function value is the first choice of where\n");
  printf("to put the items in the table. For this test, the probe value (using the\n");
  printf("g() function) is the second choice if there is a collision.\n");
  print_table(1, table_size, table, num_keys);                /* print table 2 */

  printf(newline);
  for(k=0; k<20; k++) printf("****");
  printf("\n\nTEST 3: Open addressing with chaining\n");
  printf("hash function: h(xyz) = (ord(x) + ord(z)) mod %d\n", table_size);
  printf("linear probe : (i+1) mod %d, (i+2) mod %d, ... (i+ %d) mod %d", m, m, m, m);
  printf("NOTE: that the first hash function value is the first choice of where\n");
  printf("to put the items in the table. For this test, the linear probe value is the\n");
  printf("second choice if there is a collision.\n");
  print_table(2, table_size, table, num_keys);                 /* print table 3 */
  
} /* end main */

/************************************* end main program *******************************/

/*
**
***
****
*****/
/*********************************** functions start below here **************************/
/*****
*****
***
**
*/

/******************************** instructions() *********************************/
/*
*  PURPOSE: to give user instructions
*  INPUT  : none
*  OUTPUT : text to standard output
*
*/

void instructions()
{
 printf("This program tests various hashing strategies. Input data consists of three\n");
 printf("letter month abbreviations. Since we want to test various strategies, this\n");
 printf("program has been implemented as a multi-dimensional array so that all strategies\n");
 printf("can be calculated at once and printed out in tables for your perusal. You will\n");
 printf("be asked what size table and how many keys to use. Sometimes a larger table size\n");
 printf("enable less collisions to occur, but this is not always the case for all\n");
 printf("hashing strategies. This demonstration will use the twelve month's of the year\n");
 printf("for input, and will not give you any output until all input is entered  unless\n");
 printf("there is a FATAL ERROR! which means there was an unresolvable (within this\n");
 printf("strategy) collision amongst input. Press return when ready to run tests...");
 getchar();
}

/***************************** end instructions() *********************************/


/******************************** collision() *********************************/
/*
*  PURPOSE: to signal if a collision has occured
*  INPUT  : an object of hash_type
*  OUTPUT : returns TRUE or FALSE
*
*/

int collision(table)
hash_type table;
{
 if(table.link == -1) return(FALSE);
  else return(TRUE);
}

/***************************** end collision() *********************************/


/******************************** hash() *********************************/
/*
*  PURPOSE   : to calculate a hash value for a character string
*  INPUT     : ptr, pointer to character string and modulus desired in calculation
*  OUTPUT    : returns hash value
*              hash(xyz) = (ord(x) + ord(z)) mod modulus
*              where the string pointed to by ptr is xyz
*
*/

int hash(ptr, modulus)
char *ptr;
int modulus;
{
 return( (*ptr + *(ptr+2)) mod modulus );
}

/*************************** end hash() *******************************/


/******************************** doub_hash() *********************************/
/*
*  PURPOSE: calculates a modified linear probe value using a hash function
*           and an incremental value relating to how many times we've tried
*           to find a free location
*  INPUT  : place, location where collision occurred
*           ptr, pointer to character string
*           modulus, modulus for calculation
*           next_incr, 1 the first time, 2 the 2nd time, n the nth time
*             this calculation needs to be done for a particular object
*  OUTPUT : returns value calculated
*
*/

int doub_hash(place, ptr, modulus, next_incr)
int place;
char *ptr;
int modulus;
int next_incr;
{
 return( (place + next_incr*gash(ptr, modulus)) mod modulus );
}

/******************************** end doub_hash() ***********************/


/******************************** gash() *********************************/
/*
*  PURPOSE   : to calculate a hash value for a character string
*  INPUT     : ptr, pointer to character string and modulus desired in calculation
*  OUTPUT    : returns hash value
*              gash(xyz) = (ord(x) + ord(y)) mod modulus
*              where the string pointed to by ptr is xyz
*
*/

int gash(ptr, modulus)
char *ptr;
int modulus;
{
 return( (*ptr + *(ptr+1)) mod modulus );
}

/*************************** end gash() *******************************/


/******************************** lin_probe() *********************************/
/*
*  PURPOSE: to find an open spot in table to place an object
*  INPUT  : place, location where collision occurred
*           modulus, desired modulus for calculation
*           next_incr, 1 the first time, 2 the 2nd time, n the nth time that
*             this function is called to try to locate a spot for a particular object
*  OUTPUT : returns value calculated
*
*/

int lin_probe(place, modulus, next_incr)
int place;
int modulus;
int next_incr;
{
 return( (place+next_incr) mod modulus );
}

/******************************** end lin_probe() ***********************/
 

/******************************** test() *********************************/
/*
*  PURPOSE: performs the specific test for the specific table being used
*  INPUT  : page, what page do we use for variables (what test are we doing?)
*           place, current location in table
*           input, user input with associated values
*           size, of the table we're working with
*           index, how many times have tried to place this object
*  OUTPUT : performs proper test determined by what 'page' we are on; modifies counts
*           associated with object
*  NOTE: see hash_type.h for structure of object; if you want to add a new test,
*        add another case statement in this routine.
*         This looks much more complex than it is with all the indexes to keep track of
*        which array we are presently using.
*
*
*/

void test(page, place, table, input, size, index)
int page, place[NUM_TESTS], size, index;
hash_type table[NUM_TESTS][MAX_TABLE], input[NUM_TESTS];
{
 int p = 0;     /* p, */
 int r = 0;    /* r, */
 int s = 0;   /* s: indices */
 int temp;   /* temporary placeholder */
 
 table[page][place[page]].crash += 1;        /* increment collision count */

 switch(page)
  {
   case 0:  place[page]          = lin_probe(place[page], size, index);
            input[page].lin_prob = place[page];
            break;
   case 1:  place[page]       = doub_hash(place[page], input[page].key, size, index);
            input[page].doubl = place[page];
            break;
   case 2:  if(table[page][place[page]].link <= 0)                                /* this is the last entry on chain */
             {
              temp = place[page];                                                /* save location of this last item in chain */
              while( collision(table[page][place[page]]) && (++p < size) )     /* while collisions */
               {
                place[page]          = lin_probe(place[page], size, p);       /* use linear probe */
                input[page].lin_prob = place[page];                          /* to place item */
               } /* end while */

               table[page][temp].link = place[page];                       /* and link prior last chain item */
                                                                          /* to this new last chain item */
             }      /* end if */
             else               /* not last in chain but collisions */
              {
               while( (table[page][place[page]].link > 0) && (++r < size) )
                 place[page] = table[page][place[page]].link;
                                                       /* link <= 0 now */
               temp = place[page];                    /* i.e. last in chain */
               while( collision(table[page][place[page]]) && (++s < size) )
                {
                 place[page]          = lin_probe(place[page], size, s);
                 input[page].lin_prob = place[page];
                } /* end while */

                table[page][temp].link = place[page];

               } /* end else */

            break;
   default: printf("Illegal argument to this (test()) function!\n");
            break;
  }
}

/************************************ end test() *********************************/


/******************************** new_spot() *********************************/
/*
*  PURPOSE: places object in table
*  INPUT  : page, what page we are on; which table to put object in
*           place, what place to put it in
*           table, array starting address
*           input, the object
*  OUTPUT : places object in its new spot in table
*  NOTE: if new tests are added, the case statement here will need to be expanded
*
*/

void new_spot(page, place, table, input)
int page, place[NUM_TESTS];
hash_type table[NUM_TESTS][MAX_TABLE], input[NUM_TESTS];
{
 sprintf(table[page][place[page]].key, "%s", input[page].key);
 table[page][place[page]].hash     = input[page].hash;
 table[page][place[page]].link     = -2;

 switch(page)
  {
   case 0:                 /* fall through and do same as for page 2 */
   case 2:  table[page][place[page]].lin_prob = input[page].lin_prob;
            break;
   case 1:  table[page][place[page]].doubl = input[page].doubl;
            break;
   default: printf("Illegal argument for this (new_spot()) function!\n");
            break;
  } /* end switch */
}

/************************************** end new_spot() *************************************/
 

/******************************** print_table() *********************************/
/*
*  PURPOSE: to print out the table of values and information
*  INPUT  : page, which page we're on
*           table_size, of the table
*           num_keys, number of keys used
*           table, start address of the array
*  OUTPUT :
*
*/

void print_table(page, table_size, table, num_keys)
int page;
int table_size;
int num_keys;
hash_type table[NUM_TESTS][MAX_TABLE];
{
 int index;
 int total = 0;   /* to calculate total collisions */

 printf("\nNOTE: that missing months are homeless months that couldn't find an empty table\n");
 printf("location. Disregard minus (-) values in the hash, linear probe, or link columns.\n");
 printf("(-) is a marker value. EXCEPTION: For TEST 3, -2 in the chain link field means\n");
 printf("this is the last chain item.\n");
 printf("\ntable  # collisions            first                    double     chain link\n");
 printf("index  on this index  month  hash value  linear probe  hash value    value\n");
 for(index=0; index<20; index++) printf("----");
 for(index=0; index<table_size; index++)
  printf("\n %3d       %3d         %s      %3d           %3d         %3d         %3d",\
    index, table[page][index].crash, table[page][index].key, table[page][index].hash, table[page][index].lin_prob, table[page][index].doubl, table[page][index].link);

 for(index=0; index<table_size; index++) total += table[page][index].crash;

 printf("\n\n          ===> Total collisions using this strategy was %d <===\n", total);
     printf("          ===>    using a table size of %d and %d keys     <===\n\n", table_size, num_keys);
}

/**************************** end print_table() ******************************/


#ifdef nothing   /* NOTE!!! this is the include file hash.h placed here so */
                /* there's one less file to get lost on the BBS */
 
typedef struct hash_type
{
 char *key;                /* string */
 char array[4];           /* useage: key = array; 3 char array + terminating '\0' */

 int link;              /* will be used in chaining */
 int crash;            /* used as boolean */
 int hash;            /* save hash value */
 int doubl;          /* save 2nd hash value */
 int lin_prob;      /* save linear probe value */
} hash_type;

#endif
 
 
